%% Überbestimmtes lineares Gleichungssystem
% Für ein System aus 3 Gleichungen in zwei Unbekannten sind hier geometrische 
% Interpretationen und Lösungsverfahren illustriert. CB SS2021

clear

% Das ist das Gleichungsystem in schulmäßiger Schreibweise
%    64 x + 255 y = 112
%   224 x +  20 y = 128
%   208 x + 147 y = 144
%
% Woher kommt dieses System? Es ergibt sich aus der Fragestellung (siehe
% Vorlesungsfolien): Wie gut lässt sich der Farbton SlateGray (RGB 112 128
% 144) aus Turquoise (RGB 64 224 208) und DeepPink (RGB 255  20 147)
% zusammenmischen?

% Das sind Matrix A und rechte Seite b des Systems. 
A = [ 64  255 
     224   20 
     208  147 ];
b = [112;128;144];

% MATLABs Bergabstrich-Operator berechnet x_lsq (lsq steht für least squares)
x_lsq = A\b

% Wir setzen das von MATLAB mit A\b berechnete
% Ergebnis in die Gleichung ein und sehen nach, ob die richtige rechte
% Seite herauskommt)
b_lsq = A*x_lsq;

% Die Abweichung vom gewünschten Ergebnis b ist das Residuum r = b - b_lsq
r = b - b_lsq

% Der Resuiduums-Vektor r ist nicht 0, das gewünschten Ergebnis lässt sich
% nicht erreichen, das Gleichungssystem ist somit im exakten Sinn nicht
% lösbar. Die kleinste-Quadrate-Lösung x_lsq ist die "am wenigsten falsche"
% Kompromisslösung: Die Summe der Quadrate im Residuumsvektor ist minimal. 
norm(r)
%% Sehr naive Lösungsansatz: überzählige Gleichungen weglassen
% Man könnte doch einfach die letzte Gleichung weglassen, dann hätte man ein 
% eindeutig lösbares System aus zwei Gleichungen und Unbekannten: 
% 64 x + 255 y = 112 
% 224 x + 20 y = 128
% 
% falsch! Falsch!! FALSCH!!! 

x_falsch = A([1 2],[1 2])\b([1 2])
r = b - A*x_falsch

% Im Restvektor sind die ersten beiden Komponenten 0 (eh klar, weil die
% ersten beiden Gleichungen exakt gelöst wurden), aber die dritte
% Komponente ist groß. Länge der Restvektors ist größer als vorher:
norm(r)
%% Geometrische Interpretation 1: Schnittpunkt von Geraden im R2
% Die drei Zeilen des Gleichungssystems entsprechen drei Geradengleichungen 
% 64 x + 255 y = 112 
% 224 x + 20 y = 128 
% 208 x + 147 y = 144
% 
% diese drei Geraden sind hier gezeichnet. (Die Geradengleichungen müssen dazu 
% jeweils auf die Form y = k x + d gebracht werden, das sieht hier unten unübersichtlich 
% aus, ist aber einfache Mittelschulalgebra.

x=[0.3 0.8];
y1 =( b(1)-A(1,1)*x)/A(1,2);
y2 =( b(2)-A(2,1)*x)/A(2,2);
y3 =( b(3)-A(3,1)*x)/A(3,2);

plot(x,y1,x,y2,x,y3)
grid on

% Die drei Geraden haben keinen gemeinsamen Schnittpunkt. Das oben berechnete Ergebnis
% x_falsch entspricht übrigens dem Schnittpunkt der roten mit der blauen
% Geraden.
% Sie können aus der Graphik (x,y)-Werte "in der Mitte des Fehlerdreiecks"
% auslesen und probieren, wie gut die Gleichung erfüllt ist. Probieren Sie
% hier aus:
r = A*[0.54; 0.26] - b
norm(r)

%%
plot(x,y1,x,y2,x,y3,x_lsq(1),x_lsq(2),'ok') % Gerade und lsqr-Punkt

%% Geometrische Interpretation 2: Linearkombination der Spaltenvektoren im R3
% Eine Matrix-Vektor-Multiplikation lässt sich auch als Linearkombination
% der Spaltenvektoren interpretieren. Das Gleichungssystem in dieser
% Schreibweise:
%
%   [ 64 ]     [255]     [112]
%   [ 224] x + [ 20] y = [128]
%   [ 208]     [147]     [144]
% 
% hier sind die beiden Spaltenvektoren und die von ihnen aufgespannte Ebene
% dargestellt. Auch der Zielpunkt (Vektor auf der rechten Seite) ist
% eingezeichnet. Er liegt fast, aber eben nur fast auf der Ebene. Das
% heißt, er lässt sich nicht als Linearkombination der Spaltenvektoren
% erreichen. (Drehen Sie und zoomen Sie in die Darstellung! )
%
% Der Punkt auf der Ebene mit geringstem Abstand zum Zielpunkt ist jener,
% für den der Abstandsvektor normal auf die Ebene steht. Der Normalabstand
% ist genau die Länge des Restvektors bei der kleinsten-Quadrate-Lösung.
% 

openfig('UebbGlsysBeispFig1.fig')
% (Das Erstellen dieser Abbildung erfordert etwas umständlichen Code, der
% ist hier nicht enthalten, um diese Beispieldatei nicht mit unnötigen
% Details zu überladen...)

%% Lösungsweg über Normalengleichungen
%
ATA = A'*A
ATb = A'*b

% Das ist ein 2x2-Gleichungesystem, seine Lösung ist die
% kleinste-Quadrate-Näherungslösung des überbestimmten Systems

x_normgl = ATA\ATb 

%% Lösungsweg über QR-Zerlegung
% Diesen Rechenweg führt MATLABs Bergabstrich-Operator durch. Hier sind die
% Einzelschritte angeschrieben; damit lassen sich die Zahlenwerte auf den
% Vorlesungsfolien nachprüfen.
[Q, R] = qr(A)
QTb = Q'*b

% Im Gleichungssystem R*x = Q'*b lautet die letzte Gleichung 0 = 10; die
% Komponente QTb(3) ist wiederum der kleinstmögliche Restfehler.
% Man kann sagen: Drehung mit Q^T transformiert das überbestimmte System in
% zwei exakt lösbare Gleichungen und einen unlösbaren Rest. 
% 
% Es ist ganz wesentlich, bevor man die dritte Gleichung  weglässt, das Gleichungssystem  mittels Q^T zu
% drehen; tut man das nicht, macht man den oben beschriebenen sehr naiven
% und sehr suboptimalen Lösungsversuch.

%% Lösungsweg über Singulärwertzerlegung
[U, S, V] = svd(A)
UTb = U'*b
y = S\UTb
V*y

%% Die Pseudoinverse
% Die Rechenschritte beim Aufstellen und Lösen der Normalengleichungen
% lassen sich auch so formulieren:
ATA = A'*A
Api = inv(ATA)*A'
x_pi = Api*b

% MATLAB hat einen eigenen Befehl zur Berechnung der Pseudoinversen: 
pinv(A)

%% Was noch alles schiefgehen kann:
% Die Methode der Normalengleichungen versagt, wenn die Matrix A^T*A
% singulär ist. 

% Hier ist eine andere Beispielmatrix:

A1= [ 64   32
      224  112
      176   88 ]; 
  
% In diesem Fall ist das Originalsystem überbestimmt, abed das System der
% Normalengleichungen unterbestimmt! Das heisst, es gibt keine exakte
% Lösung des Originalsystems, aber unendlich viele Lösungen mit
% kleinstem-Quadrate-Restvektor. 

% Normalengl.
x_ngl=(A1'*A1)\A1'*b 

% QR, Bergabstrich
x_qr= A1\b

% pinv
x_pi= pinv(A1)*b

% SVD
[U, S, V] = svd(A1)
x_svd=V*(S\(U'*b))
 
 






