%% Experimente zur SVD
clear
%% Datenmatrix 
% Siehe Beispiel auf den Vorlesungsfolien: Datenmatrix Bewertung von Filmen
% durch Filmfans: 11 Filme (Zeilen) 3 Fans (Spalten)

A = [2, 1, 5, 5, 5, 0, 0, 0, 0, 0, 0
     1.5, 1.5, 3, 3, 3.5, 1, 1, 1, 1, 1, 0.5
     1, 4, 0, 0, 0, 5, 5, 5, 4, 5, 3]';
 
plot3(A(:,1),A(:,2),A(:,3),'.','MarkerSize',25)
xlim([0,5])
ylim([0,5])
zlim([0,5])

%% SVD (economy size form)
% die "Sparvariante" der SVD lässt  in U oder V jene Spalten oder Zeilen weg,
% die nichts beitragen, weil S sie mit 0 multiplizieren würde. 
% 
 [U, S, V]=svd(A,'econ')

%% Eigenschaften von U und V
% orthogonale Spaltenvektoren (aber Zeilen sind in der "Sparvariante" 
% nicht orthogonal!)
U' * U  
V' * V

A0 = U*S*V' % Originalmatrix laesst sich so rekonstruieren
norm(A)
% Invarianz der Norm: 2-Norm von A ist 2-Norm von S, zugleich größter Singulärwert
norm(S) 

%% Rank 1 - Rekonstruktion
A1= U(:,1)*S(1,1)*V(:,1)'
norm(A1-A)

%% Rank 2
A2= A1 + U(:,2)*S(2,2)*V(:,2)'
norm(A2-A)
S(3,3) % um genau dieser Wert unterscheiden sich A2 und A in der 2-Norm

%% Rank 3
A3= A2 + U(:,3)*S(3,3)*V(:,3)'
norm(A3-A) % das ist wieder die Originalmatrix



%% Projektion auf den Unterraum: Ausgleichsebene
% Hier erkennt man: die Datenpunkte liegen "nicht wirklich" in einem
% dreidimensionalen Raum, slie leigen alle (fast) auf einer Ebene, also in
% einem zweidimensionalen Unterraum. Jeder Datenpunkt lässt sich durch zwei
% Koordinaten definieren, drei Koordinaten sind unnötig.

mn = mean(A) 
A = A - mean(A);  % stelle auf Mittelwert 0 ein: Schwerpunkt der Daten im Ursprung
[U, S, V]=svd(A,'econ')

hold on
% plot3(A0(1,:),A0(2,:),A0(3,:),'.g','MarkerSize',25)
d =mn*V(:,3);
[X,Y]=meshgrid(0:0.4:5,0:0.4:5); 
Z = (d  - X*V(1,3) - Y*V(2,3))/V(3,3);
surf(X,Y,Z,'FaceAlpha',0.1)
hold off




%% Compression Example
% (this is based on material I got from somewhere I don't remember...)
%
% We now load a sample photograph into MATLAB with *|imread|* and check its
% size (we will display the photograph later, after converting the data to
% shades of gray): 

%photoarray = imread('sample1.jpg');
photoarray = imread('katze2.jpg');
size(photoarray)

%%
% The values in the array are red, green, and blue intensities (integer
% values from 0 to 255) for each of 1536*2048 pixels.
%
% Next, we separate the colors and convert to floating-point values between
% 0 and 1.  We use single precision because it is sufficient for our
% calculations, and our objective is to store efficiently the essential
% data in the image.  (We are still converting each byte of the integer
% image data into four bytes of floating-point data, but the floating-point
% is necessary for the calculations we do below.)
redpix = single(photoarray(:,:,1))/255;
greenpix = single(photoarray(:,:,2))/255;
bluepix = single(photoarray(:,:,3))/255;

%%
% To illustrate the ideas below more simply, we convert to a
% black-and-white (more precisely, ``grayscale'') image.
% The following is a standard formula to convert RGB intensities to
% gray intensities:
graypix = 0.299*redpix + 0.587*greenpix + 0.114*bluepix;

%%
% Since we have converted to floating-point values, we use *|imagesc|* rather
% than *|image|* to display the photograph. 
% We use |colormap|
% to get a grayscale display, and *|axis|* to turn off axis labels and
% ensure the correct aspect ratio:
figure(1)
imagesc(graypix)
colormap(gray), axis off image

%%
% Now we compute the singular value decomposition of the grayscale data.
[U,S,V] = svd(graypix);

%%
% To see how fast they decay, we graph the first 100 singular values.  We
% open a new figure window so that the photograph remains visible if this
% M-file is used interactively:
figure(2)
plot(diag(S(1:100,1:100)))

%%
% The following function computes and displays the rank $n$ approximation
% to our image:
showrank = @(n) imagesc(U(:,1:n)*S(1:n,1:n)*V(:,1:n)');

%%
% Below we show the approximations of rank 1, 3, 10, and 30:
figure(3)
colormap(gray)
rankvals = [1 3 10 30];
set(0,'DefaultAxesFontSize', 10)
for m=1:4
    subplot(2,2,m)
    showrank(rankvals(m)), axis off image
    title(['Rank ' num2str(rankvals(m))])
end
set(0,'DefaultAxesFontSize', 16)

%%
% A rank 1 matrix can only produce a ``plaid'' image, but by the time we get
% to rank 10, the image starts to look like a photograph, and by rank 30 it
% is clear we are looking at a landscape.  Next, let's compare the rank 100
% approximation to the full-resolution image.  We clear the existing figure
% so that the new image is not shown in a subplot:
figure(4)
clf
showrank(100); axis off image
colormap(gray), 
%%
% The rank 100 approximation is not as sharp as the original, but the loss
% of quality might not be noticed if the original were not available.
% As a rough comparison to JPEG compression, let's store the grayscale
% image as a JPEG file with default parameters and see how big it is:

imwrite(graypix, 'graysamp.jpg', 'jpg')
d = dir('graysamp.jpg'); d.bytes

%%
% The raw image data consist of $1536 \cdot 2048$ intensities that could
% be stored as 1 byte apiece, for a total of 3 megabytes.  Since our rank
% $r$ approximation requires $(1536 + 2048)r = 3584r$
% floating point numbers, and each of our numbers is stored in single
% precision (4 bytes), we can compute as follows the rank for which our
% approximation would require about the same amount of data as the JPEG
% file:

r = d.bytes/3584/4





